import random
import unittest
from typing import Optional, Union, Dict, Any, Tuple
from uuid import uuid4

from StructNoSQL import DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiCachingTable, FieldSetter, \
    FieldGetter, FieldRemover
from StructNoSQL.clients_middlewares.external_dynamodb_api.external_dynamodb_api_basic_table import ExternalDynamoDBApiBasicTable
from tests.components.playground_table_clients import TEST_ACCOUNT_ID


def test_get_simple_float_field(
        self: unittest.TestCase,
        table_client: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    simple_random_float_value: float = random.randint(1, 10000) / 100
    simple_float_update_success: bool = table_client.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleFloatField', value_to_set=simple_random_float_value
    )
    self.assertTrue(simple_float_update_success)
    if is_caching is True:
        self.assertTrue(table_client.commit_operations())
        table_client.clear_cached_data()

    retrieved_float_value: Optional[float] = table_client.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleFloatField'
    )
    self.assertEqual((
        simple_random_float_value if is_caching is not True else
        {'value': simple_random_float_value, 'fromCache': False}
    ), retrieved_float_value)


def test_get_multiple_float_fields(
        self: unittest.TestCase,
        table_client: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    container_field_one_random_text_value: float = random.randint(1, 10000) / 100
    container_field_two_random_text_value: float = random.randint(10001, 20000) / 100
    first_table_container_fields_update_success: bool = table_client.update_multiple_fields(
        key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='floatsContainer.floatOne', value_to_set=container_field_one_random_text_value),
            FieldSetter(field_path='floatsContainer.floatTwo', value_to_set=container_field_two_random_text_value)
        ]
    )
    self.assertTrue(first_table_container_fields_update_success)

    if is_caching is True:
        self.assertTrue(table_client.commit_operations())
        table_client.clear_cached_data()

    second_table_retrieved_container_fields_without_data_validation: Dict[str, Optional[Any]] = table_client.get_multiple_fields(
        key_value=TEST_ACCOUNT_ID, getters={
            'one': FieldGetter(field_path='floatsContainer.floatOne'),
            'two': FieldGetter(field_path='floatsContainer.floatTwo'),
        }
    )
    self.assertEqual({
        'one': (
            container_field_one_random_text_value if is_caching is not True else
            {'value': container_field_one_random_text_value, 'fromCache': False}
        ),
        'two': (
            container_field_two_random_text_value if is_caching is not True else
            {'value': container_field_two_random_text_value, 'fromCache': False}
        )
    }, second_table_retrieved_container_fields_without_data_validation)


def test_remove_simple_float_field(
        self: unittest.TestCase,
        table_client: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    simple_random_float_value: float = random.randint(1, 10000) / 100
    simple_float_field_update_success: bool = table_client.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleFloatField', value_to_set=simple_random_float_value
    )
    self.assertTrue(simple_float_field_update_success)
    if is_caching is True:
        self.assertTrue(table_client.commit_operations())
        table_client.clear_cached_data()

    removed_simple_float_field_value: Optional[float] = table_client.remove_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleFloatField'
    )
    self.assertEqual((
        simple_random_float_value if is_caching is not True else
        {'value': simple_random_float_value, 'fromCache': False}
    ), removed_simple_float_field_value)



import unittest

from tests.components.playground_table_clients import PlaygroundInoftVocalEngineCachingTable
from tests.tests_caching_table.table_models import ExternalDynamoDBApiTableModel


class TestExternalDynamoDBApiCachingTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundInoftVocalEngineCachingTable(data_model=ExternalDynamoDBApiTableModel)
        self.users_table.debug = True

    def test_simple_get_field(self):
        from tests.tests_caching_table.cases_shared import test_simple_get_field
        test_simple_get_field(self, users_table=self.users_table)

    def test_set_then_get_field_from_cache(self):
        from tests.tests_caching_table.cases_shared import test_set_then_get_field_from_cache
        test_set_then_get_field_from_cache(self, users_table=self.users_table)

    def test_set_then_get_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_set_then_get_multiple_fields
        test_set_then_get_multiple_fields(self, users_table=self.users_table)

    def test_set_then_get_pack_values_with_one_of_them_present_in_cache(self):
        from tests.tests_caching_table.cases_shared import test_set_then_get_pack_values_with_one_of_them_present_in_cache
        test_set_then_get_pack_values_with_one_of_them_present_in_cache(self, users_table=self.users_table)

    def test_debug_simple_set_commit_then_get_field_from_database(self):
        from tests.tests_caching_table.cases_shared import test_debug_simple_set_commit_then_get_field_from_database
        test_debug_simple_set_commit_then_get_field_from_database(self, users_table=self.users_table)

    def test_update_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_update_multiple_fields
        test_update_multiple_fields(self, users_table=self.users_table)

    def test_set_delete_field(self):
        from tests.tests_caching_table.cases_shared import test_set_delete_field
        test_set_delete_field(self, users_table=self.users_table)

    def test_set_remove_field(self):
        from tests.tests_caching_table.cases_shared import test_set_remove_field
        test_set_remove_field(self, users_table=self.users_table)

    def test_dict_data_unpacking(self):
        from tests.tests_caching_table.cases_shared import test_dict_data_unpacking
        test_dict_data_unpacking(self, users_table=self.users_table)

    def test_list_data_unpacking(self):
        from tests.tests_caching_table.cases_shared import test_list_data_unpacking
        test_list_data_unpacking(self, users_table=self.users_table)

    def test_set_remove_multi_selector_field_and_field_unpacking(self):
        from tests.tests_caching_table.cases_shared import test_set_remove_multi_selector_field_and_field_unpacking
        test_set_remove_multi_selector_field_and_field_unpacking(self, users_table=self.users_table)

    def test_set_delete_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_set_delete_multiple_fields
        test_set_delete_multiple_fields(self, users_table=self.users_table)

    def test_set_remove_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_set_remove_multiple_fields
        test_set_remove_multiple_fields(self, users_table=self.users_table)

    def test_set_get_fields_with_primary_index(self):
        from tests.tests_query_operations.cases_shared import test_set_get_fields_with_primary_index
        test_set_get_fields_with_primary_index(self, users_table=self.users_table, primary_key_name='accountProjectTableKeyId', is_caching=True)


if __name__ == '__main__':
    unittest.main()



import unittest
import uuid
from typing import Optional

from StructNoSQL import TableDataModel, BaseField
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    username = BaseField(field_type=str, required=False)

class TestsUpdateNonExistingRecord(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_update_non_existing_record(self):
        random_record_id: str = f"randomRecordId_{uuid.uuid4()}"

        update_success: bool = self.users_table.update_field(
            key_value=random_record_id, field_path='username', value_to_set="Robinson"
        )
        self.assertTrue(update_success)

        retrieved_username: Optional[str] = self.users_table.get_field(
            key_value=random_record_id, field_path='username'
        )
        self.assertEqual("Robinson", retrieved_username)

        record_deletion_success: bool = self.users_table.delete_record({
            'accountId': random_record_id
        })
        self.assertTrue(record_deletion_success)

if __name__ == '__main__':
    unittest.main()



import unittest

from tests.components.playground_table_clients import PlaygroundDynamoDBCachingTable
from tests.tests_update_field_return_old.table_models import DynamoDBTableModel


class TestsDynamoDBCachingTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBCachingTable(data_model=DynamoDBTableModel)
        self.users_table.debug = True

        self.DYNAMODB_CASE_KWARGS = {'self': self, 'users_table': self.users_table, 'is_caching': True}
        self.SHARED_CASE_KWARGS = {**self.DYNAMODB_CASE_KWARGS, 'primary_key_name': 'accountId'}

    def test_update_field_return_old(self):
        from tests.tests_update_field_return_old.cases_shared import test_update_field_return_old
        test_update_field_return_old(**self.SHARED_CASE_KWARGS)

    def test_update_multiple_fields_return_old(self):
        from tests.tests_update_field_return_old.cases_shared import test_update_multiple_fields_return_old
        test_update_multiple_fields_return_old(**self.SHARED_CASE_KWARGS)



from setuptools import setup, find_packages

setup(
    name="StructNoSQL",
    version="2.6.4",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pydantic", "boto3", "requests"],
    url="https://github.com/Robinson04/StructNoSQL",
    license="MIT",
    author="Inoft",
    author_email="robinson@inoft.com",
    description="Structured document based NoSQL client for DynamoDB with automatic data validation and advanced database queries functions.",
)



import unittest

from tests.components.playground_table_clients import PlaygroundDynamoDBCachingTable
from tests.tests_query_operations.table_models import DynamoDBTableModel


class TestsDynamoDBCachingTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBCachingTable(data_model=DynamoDBTableModel)
        self.users_table.debug = True

        self.DYNAMODB_CASE_KWARGS = {'self': self, 'users_table': self.users_table, 'is_caching': True}
        self.SHARED_CASE_KWARGS = {**self.DYNAMODB_CASE_KWARGS, 'primary_key_name': 'accountId'}

    def test_set_get_fields_with_primary_index(self):
        from tests.tests_query_operations.cases_shared import test_set_get_fields_with_primary_index
        test_set_get_fields_with_primary_index(**self.SHARED_CASE_KWARGS)

    def test_set_get_paginated_fields_with_primary_index(self):
        from tests.tests_query_operations.cases_shared import test_set_get_paginated_fields_with_primary_index
        test_set_get_paginated_fields_with_primary_index(**self.SHARED_CASE_KWARGS)

    def test_set_get_fields_with_secondary_index(self):
        from tests.tests_query_operations.cases_dynamodb import test_set_get_fields_with_secondary_index
        test_set_get_fields_with_secondary_index(**self.DYNAMODB_CASE_KWARGS)

    def test_set_get_fields_with_overriding_names(self):
        from tests.tests_query_operations.cases_dynamodb import test_set_get_fields_with_overriding_names
        test_set_get_fields_with_overriding_names(**self.DYNAMODB_CASE_KWARGS)

    def test_set_get_fields_with_multi_selectors(self):
        from tests.tests_query_operations.cases_dynamodb import test_set_get_fields_with_multi_selectors
        test_set_get_fields_with_multi_selectors(**self.DYNAMODB_CASE_KWARGS)

    def test_set_get_multiple_fields_with_special_inner_keys(self):
        from tests.tests_query_operations.cases_dynamodb import test_set_get_multiple_fields_with_special_inner_keys
        test_set_get_multiple_fields_with_special_inner_keys(**self.DYNAMODB_CASE_KWARGS)


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import Dict
from uuid import uuid4

from StructNoSQL import FieldGetter, FieldSetter, BaseField, MapModel, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    fieldAlreadyInitializedRequiringNewValue = BaseField(field_type=str, required=False)
    class ContainerModel(MapModel):
        firstNestedValue = BaseField(field_type=str, required=False)
        class NestedContainerModel(MapModel):
            secondNestedValue = BaseField(field_type=str, required=False)
        nestedContainer = BaseField(field_type=Dict[str, NestedContainerModel], required=False)
    containerThatWillFailRequestAndRequireFieldsInitialization = BaseField(field_type=Dict[str, ContainerModel], key_name='id', required=False)

class TestRequestFailsOnFieldsInitializationButRequireExistingFieldsToBeUpdateD(unittest.TestCase):
    """
    This test a rare case, where if the update_multiple_fields is used, and one of the field to update caused the
    request to failed (for example, trying to navigate into an item that does not exist), the library will of course
    try to initialize the missing fields. The library has no idea which field caused the request to fail, so, it will
    try to initialize all of the field, one per one, only if they are not existing. In a single update operation,
    this never cause issues, but in a multi fields update operation, we could have an existing field, that did not
    cause the operation the fail, and that would need to be modify. Yet, since it already exist, the library will not
    correctly modify it as the request instructed it to do. To prevent this issue, if an item already exist the library
    will retrieve its UPDATED_NEW value in the same request (to avoid a get request), and check if the value present in
    the database is the value that our request specified. If its not the case, we will send a new request to modify the
    value. This request will not try to initialize fields if she fails, to avoid potential infinite recursion.
    """

    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_with_string_value_as_target(self):
        random_id = str(uuid4())
        random_value_one = f"one_{uuid4()}"
        random_value_two = f"two_{uuid4()}"

        initial_update_with_empty_value_success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='fieldAlreadyInitializedRequiringNewValue', value_to_set="initial"
        )
        self.assertTrue(initial_update_with_empty_value_success)

        update_success = self.users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='fieldAlreadyInitializedRequiringNewValue', value_to_set=random_value_one),
            FieldSetter(
                field_path='containerThatWillFailRequestAndRequireFieldsInitialization.{{id}}.firstNestedValue',
                query_kwargs={'id': random_id}, value_to_set=random_value_two
            )
        ])
        self.assertTrue(update_success)

        get_response_data = self.users_table.get_multiple_fields(key_value=TEST_ACCOUNT_ID, getters={
            'one': FieldGetter(field_path='fieldAlreadyInitializedRequiringNewValue'),
            'two': FieldGetter(field_path='containerThatWillFailRequestAndRequireFieldsInitialization.{{id}}.firstNestedValue', query_kwargs={'id': random_id})
        })
        self.assertEqual(get_response_data.get('one', None), random_value_one)
        self.assertEqual(get_response_data.get('two', None), random_value_two)

    def test_with_dict_values_as_target(self):
        nested_container_field_path = 'containerThatWillFailRequestAndRequireFieldsInitialization.{{id}}.nestedContainer'
        random_first_container_id_one = f"first_container_{uuid4()}"
        random_first_container_id_two = f"second_container_{uuid4()}"
        random_nested_container_id_one = f"id_one_{str(uuid4())}"
        random_nested_container_id_two = f"id_two_{str(uuid4())}"
        random_container_field_value_one = {'secondNestedValue': f"container_field_value_one_{uuid4()}"}
        random_container_field_value_two = {'secondNestedValue': f"container_field_value_two_{uuid4()}"}

        # Initial update that put a single item in the container without risking to override existing fields
        initial_update_with_empty_value_success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path=nested_container_field_path + '.{{nestedContainerKey}}',
            query_kwargs={'id': random_first_container_id_one, 'nestedContainerKey': random_nested_container_id_one},
            value_to_set=random_container_field_value_one
        )
        self.assertTrue(initial_update_with_empty_value_success)

        # Multi update operation, that could cause the
        update_success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path=nested_container_field_path + '.{{nestedContainerKey}}',
            query_kwargs={'id': random_first_container_id_two, 'nestedContainerKey': random_nested_container_id_two},
            value_to_set=random_container_field_value_two
        )
        self.assertTrue(update_success)

        retrieved_containers_data = self.users_table.get_multiple_fields(key_value=TEST_ACCOUNT_ID, getters={
            'one': FieldGetter(field_path=nested_container_field_path, query_kwargs={'id': random_first_container_id_one}),
            'two': FieldGetter(field_path=nested_container_field_path, query_kwargs={'id': random_first_container_id_two})
        })
        self.assertEqual(retrieved_containers_data['one'].get(random_nested_container_id_one, None), random_container_field_value_one)
        self.assertEqual(retrieved_containers_data['two'].get(random_nested_container_id_two, None), random_container_field_value_two)


if __name__ == '__main__':
    unittest.main()



from typing import Callable, List, Dict, Optional, Any, Tuple, Iterable

from StructNoSQL import BaseField
from StructNoSQL.models import DatabasePathElement, FieldGetter, QueryMetadata
from StructNoSQL.practical_logger import message_with_vars
from StructNoSQL.tables.base_table import FieldsSwitch
from StructNoSQL.utils.data_processing import navigate_into_data_with_field_path_elements
from StructNoSQL.utils.process_render_fields_paths import process_and_make_single_rendered_database_path


def unpack_validate_retrieved_field(
        record_attributes: dict, target_field_container: Tuple[BaseField, List[DatabasePathElement]],
) -> Optional[Any]:
    item_field_object, item_field_path_elements = target_field_container
    item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
        data=record_attributes, field_path_elements=item_field_path_elements,
        num_keys_to_navigation_into=len(item_field_path_elements)
    )
    item_field_object.populate(value=item_data)
    validated_data, is_valid = item_field_object.validate_data()
    return validated_data

def unpack_validate_retrieved_field_if_need_to(
        target_field_container: Tuple[BaseField, List[DatabasePathElement]],
        data_validation: bool, record_attributes: dict,
        item_mutator: Optional[Callable[[Any, List[DatabasePathElement]], Any]] = lambda item_value, _: item_value
) -> Optional[Any]:
    field_object, field_path_elements = target_field_container
    if data_validation is True:
        return item_mutator(unpack_validate_retrieved_field(
            record_attributes=record_attributes, 
            target_field_container=target_field_container
        ), field_path_elements)
    else:
        return item_mutator(navigate_into_data_with_field_path_elements(
            data=record_attributes, field_path_elements=field_path_elements,
            num_keys_to_navigation_into=len(field_path_elements)
        ), field_path_elements)

def unpack_validate_multiple_retrieved_fields_if_need_to(
        target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]],
        data_validation: bool, record_attributes: dict,
        item_mutator: Optional[Callable[[Any, List[DatabasePathElement]], Any]] = lambda item_value, _: item_value,
        base_output_values: Optional[Dict[str, Any]] = None
):
    output_values: Dict[str, Any] = base_output_values or {}
    if data_validation is True:
        for item_key, item_container in target_fields_containers.items():
            item_field_object, item_field_path_elements = item_container

            output_values[item_key] = item_mutator(unpack_validate_retrieved_field(
                record_attributes=record_attributes, target_field_container=item_container
            ), item_field_path_elements)
    else:
        for item_key, item_container in target_fields_containers.items():
            item_field_object, item_field_path_elements = item_container

            output_values[item_key] = item_mutator(navigate_into_data_with_field_path_elements(
                data=record_attributes, field_path_elements=item_field_path_elements,
                num_keys_to_navigation_into=len(item_field_path_elements)
            ), item_field_path_elements)
    return output_values


# todo: remove the _base_unpack_getters_response_item function
def _base_unpack_getters_response_item(
        item_mutator: Callable[[Any], Any], response_item: dict,
        single_getters_target_fields_containers: Dict[str, List[DatabasePathElement]],
        grouped_getters_target_fields_containers: Dict[str, Dict[str, List[DatabasePathElement]]]
) -> Dict[str, Any]:
    output_data: Dict[str, Any] = {}
    for item_key, item_field_path_elements in single_getters_target_fields_containers.items():
        retrieved_item_data = navigate_into_data_with_field_path_elements(
            data=response_item, field_path_elements=item_field_path_elements,
            num_keys_to_navigation_into=len(item_field_path_elements)
        )
        output_data[item_key] = item_mutator(retrieved_item_data)

    for container_key, container_items_field_path_elements in grouped_getters_target_fields_containers.items():
        container_data: Dict[str, Any] = {}
        for child_item_key, child_item_field_path_elements in container_items_field_path_elements.items():
            container_data[child_item_key] = navigate_into_data_with_field_path_elements(
                data=response_item, field_path_elements=child_item_field_path_elements,
                num_keys_to_navigation_into=len(child_item_field_path_elements)
            )
        output_data[container_key] = item_mutator(container_data)
    return output_data

# todo: make a function that validate, and one that does not, to have only one if check on validation
def _unpack_validate_getters_record_attributes_if_need_to(
        single_getters_target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]],
        grouped_getters_target_fields_containers: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]],
        item_mutator: Callable[[Any, List[DatabasePathElement]], Any], data_validation: bool,
        record_attributes: dict, base_output_values: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    output_values: Dict[str, Any] = base_output_values or {}

    for item_key, item_container in single_getters_target_fields_containers.items():
        item_field_object, item_field_path_elements = item_container
        item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
            data=record_attributes, field_path_elements=item_field_path_elements,
            num_keys_to_navigation_into=len(item_field_path_elements)
        )
        if data_validation is True:
            item_field_object.populate(value=item_data)
            validated_data, is_valid = item_field_object.validate_data()
        else:
            validated_data = item_data

        output_values[item_key] = item_mutator(validated_data, item_field_path_elements)

    for fields_container_key, fields_container_item in grouped_getters_target_fields_containers.items():
        current_container_fields_data: Dict[str, Any] = output_values.get(fields_container_key, {})

        for child_field_key, child_field_container in fields_container_item.items():
            child_field_object, child_field_path_elements = child_field_container
            item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                data=record_attributes, field_path_elements=child_field_path_elements,
                num_keys_to_navigation_into=len(child_field_path_elements)
            )
            if data_validation is True:
                child_field_object.populate(value=item_data)
                validated_data, is_valid = child_field_object.validate_data()
            else:
                validated_data = item_data

            current_container_fields_data[child_field_key] = item_mutator(validated_data, child_field_path_elements)

        output_values[fields_container_key] = current_container_fields_data
    return output_values


def _has_primary_key_in_path_elements(primary_index_name: str, fields_path_elements: Dict[str, List[DatabasePathElement]]) -> Tuple[bool, Optional[str]]:
    for client_key, item_path_elements in fields_path_elements.items():
        if len(item_path_elements) > 0:
            first_path_element = item_path_elements[0]
            if first_path_element.element_key == primary_index_name:
                return True, client_key
    return False, None

def _has_primary_key_in_path_elements_v2(primary_index_name: str, fields_path_elements: List[List[DatabasePathElement]]) -> bool:
    for item_path_elements in fields_path_elements:
        if len(item_path_elements) > 0:
            first_path_element = item_path_elements[0]
            if first_path_element.element_key == primary_index_name:
                return True
    return False

# todo: deprecate ?
def _inner_query_fields_secondary_index(
        middleware: Callable[[List[List[DatabasePathElement]]], Tuple[Optional[List[Any]], QueryMetadata]],
        primary_index_name: str, get_primary_key_database_path: Callable[[], List[DatabasePathElement]],
        fields_paths_elements: List[List[DatabasePathElement]]
) -> Tuple[Optional[dict], QueryMetadata]:
    """
    This function will force the retrieving of the primary key of the records returned by a
    secondary index query operations and retrieve a dict with the primary key of each record as key.
    """

    primary_key_is_being_requested_by_client: bool = (
        _has_primary_key_in_path_elements_v2(primary_index_name=primary_index_name, fields_path_elements=fields_paths_elements)
    )
    super_fields_path_elements: List[List[DatabasePathElement]] = (
        fields_paths_elements if primary_key_is_being_requested_by_client is True else
        [*fields_paths_elements, get_primary_key_database_path()]
    )
    retrieved_records_items_data, query_metadata = middleware(super_fields_path_elements)
    if retrieved_records_items_data is None:
        return None, query_metadata

    records_attributes: dict = {}
    for record_item_data in retrieved_records_items_data:
        # todo: stop using the primary_index_name and start using the field_name of the primary index (different)
        record_primary_key_value: Optional[Any] = record_item_data.get(primary_index_name, None)
        records_attributes[record_primary_key_value] = record_item_data
    return records_attributes, query_metadata


def _prepare_getters(fields_switch: FieldsSwitch, getters: Dict[str, FieldGetter]) -> Tuple[
    List[List[DatabasePathElement]],
    Dict[str, Tuple[BaseField, List[DatabasePathElement]]],
    Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]]
]:
    getters_database_paths: List[List[DatabasePathElement]] = []
    single_getters_target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
    grouped_getters_target_fields_containers: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]] = {}

    for getter_key, getter_item in getters.items():
        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=getter_item.field_path, fields_switch=fields_switch, query_kwargs=getter_item.query_kwargs
        )
        if is_multi_selector is not True:
            target_field_container: Tuple[BaseField, List[DatabasePathElement]]
            single_getters_target_fields_containers[getter_key] = target_field_container
            getters_database_paths.append(target_field_container[1])
        else:
            target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

            field_path_object: Dict[str, BaseField]
            getter_field_path_elements: Dict[str, List[DatabasePathElement]]

            grouped_getters_target_fields_containers[getter_key] = target_field_container
            for item_target_field_container in target_field_container.values():
                getters_database_paths.append(item_target_field_container[1])

    return getters_database_paths, single_getters_target_fields_containers, grouped_getters_target_fields_containers


def _model_contain_all_index_keys(model: Any, indexes_keys: Iterable[str]) -> bool:
    for index_key in indexes_keys:
        index_matching_field: Optional[Any] = getattr(model, index_key, None)
        if index_matching_field is None:
            print(message_with_vars(
                message="An index key selector was not found in the table model. Operation not executed.",
                vars_dict={'index_key': index_key, 'index_matching_field': index_matching_field, 'table.model': model}
            ))
            return False
    return True



import unittest
from StructNoSQL import BaseField, TableDataModel
from StructNoSQL.exceptions import InvalidFieldNameException
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable


class TestReservedChars(unittest.TestCase):
    def test_left_bracket_char(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                restrictedRightBracket = BaseField(custom_field_name='restricted_[e', field_type=str, required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)
        self.assertRaises(InvalidFieldNameException, init_table)

    def test_right_bracket_char(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                restrictedLeftBracket = BaseField(custom_field_name='restricted_]e', field_type=str, required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)
        self.assertRaises(InvalidFieldNameException, init_table)

    def test_left_curly_bracket_char(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                restrictedRightBracket = BaseField(custom_field_name='restricted_{e', field_type=str, required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)
        self.assertRaises(InvalidFieldNameException, init_table)

    def test_right_curly_bracket_char(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                restrictedLeftBracket = BaseField(custom_field_name='restricted_}e', field_type=str, required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)
        self.assertRaises(InvalidFieldNameException, init_table)

    def test_left_parenthesis_char(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                restrictedRightBracket = BaseField(custom_field_name='restricted_(e', field_type=str, required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)
        self.assertRaises(InvalidFieldNameException, init_table)

    def test_right_parenthesis_char(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                restrictedLeftBracket = BaseField(custom_field_name='restricted_)e', field_type=str, required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)
        self.assertRaises(InvalidFieldNameException, init_table)

    def test_separator_char(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                restrictedLeftBracket = BaseField(custom_field_name='restricted_|e', field_type=str, required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)
        self.assertRaises(InvalidFieldNameException, init_table)


if __name__ == '__main__':
    unittest.main()



from StructNoSQL import PrimaryIndex
from StructNoSQL.servers_middlewares.external_dynamodb_api.api_executor import ExternalDynamoDBApiExecutor


class InoftVocalEngineExternalDynamoDBApiExecutor(ExternalDynamoDBApiExecutor):
    def __init__(self):
        super().__init__(
            table_name="inoft-vocal-engine_projects-virtual-tables", region_name="eu-west-3",
            primary_index=PrimaryIndex(hash_key_name='accountProjectTableKeyId', hash_key_variable_python_type=str),
            auto_create_table=False
        )
